/*
 *	touch  Ver.0.04   Copyright (C) 2005-2011  K.Takata
 */

#ifdef UNICODE
#define PRGTYPE		"(Unicode)"
#else
#define PRGTYPE		"(ANSI)"
#endif

#define COPYRIGHT	\
	"touch " PRGTYPE "  Ver.0.04   Copyright (C) 2005-2011  K.Takata"


#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <shlwapi.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <mbstring.h>
#include <tchar.h>
#include "dirtime.h"


void usage();
int parse_reference(LPCTSTR str, SYSTEMTIME *lt);
int parse_gnu_date(LPCTSTR str, SYSTEMTIME *lt);
int parse_gnu_time(LPCTSTR str, SYSTEMTIME *lt);
int parse_date(LPCTSTR str, SYSTEMTIME *lt);
int parse_time(LPCTSTR str, SYSTEMTIME *lt);
int touch(LPCTSTR path, const FILETIME *ftCreate,
		const FILETIME *ftAccess, const FILETIME *ftModify,
		int nocreate, int recursive, int fattime);
int create_new(LPCTSTR path, const FILETIME *ftCreate,
		const FILETIME *ftAccess, const FILETIME *ftModify);
int touch_subdir(LPCTSTR path, LPCTSTR file, const FILETIME *ftCreate,
		const FILETIME *ftAccess, const FILETIME *ftModify,
		int recursive, int fattime);
BOOL SetFTime(LPCTSTR path, const FILETIME *pCreationTime,
		const FILETIME *pLastAccessTime, const FILETIME *pLastWriteTime,
		BOOL fattime);


/*
 *	main
 */
int _tmain(int argc, TCHAR *argv[])
{
	SYSTEMTIME lt; /* Local Time */
	FILETIME ft;
	FILETIME *ftAccess = NULL, *ftCreate = NULL, *ftModify = &ft;
	int ret = 0, lasterr = 0;
	int optA = 0, optC = 0, optM = 0;
	int nocreate = 0;
	int recursive = 0;
	int fattime = 0;
	
	
	if (argc < 2) {
		usage();
	}
	
	GetLocalTime(&lt);
	
	while (--argc) {
		LPTSTR p;
		
		p = *++argv;
		if (*p == '-' || *p == '/') {		/* check options */
			while (*++p) {
				switch (*p) {
				/* Archive */
				case 'a':
					/* Not Supported */
					break;
				
				/* Access Time */
				case 'A':
					optA = 1;
					break;
				
				/* Creation Time */
				case 'C':
					optC = 1;
					break;
				
				/* Modification Time */
				case 'M':
					optM = 1;
					break;
				
				/* Recursive */
				case 'R':
					recursive = 1;
					break;
				
				/* No Create (GNU touch) */
				case 'c':
					nocreate = 1;
					break;
				
				/* Reference (GNU touch) */
				case 'r':					/* -r <rfile> */
					--argc, ++argv;
					parse_reference(*argv, &lt);
					break;
				
				/* Date (GNU touch) */
				case 'd':					/* -d <string> */
					--argc, ++argv;
					parse_gnu_date(*argv, &lt);
					break;
				
				/* Time (GNU touch) */
				case 't':					/* -t [[CC]YY]MMDDhhmm[.ss[lll]] */
					--argc, ++argv;
					parse_gnu_time(*argv, &lt);
					break;
				
				/* Date */
				case 'D':					/* -D [[CC]YY-]MM-DD */
					--argc, ++argv;
					parse_date(*argv, &lt);
					break;
				
				/* Time */
				case 'T':					/* -T hh:mm[:ss[.lll]] */
					--argc, ++argv;
					parse_time(*argv, &lt);
					break;
				
				/* FAT time stamp */
				case 'F':
					fattime = 1;
					break;
				
				/* Force (BSD touch) */
				case 'f':
					/* Do Nothing */
					break;
				
				/* Quiet */
				case 'q':
					/* Not Supported */
					break;
				
				/* Help */
				case '?':					/* -h, -H, -? */
				case 'h':
				case 'H':
				default:
					usage();
					break;
				}
			}
		} else {
			ktLocalTimeToFileTime(&lt, &ft);
			if (optA) {
				ftAccess = &ft;
				ftModify = NULL;
			}
			if (optC) {
				ftCreate = &ft;
				ftModify = NULL;
			}
			if (optM) {
				ftModify = &ft;
			}
			ret = touch(*argv, ftCreate, ftAccess, ftModify,
					nocreate, recursive, fattime);
			if (ret) {
			//	return ret;
				lasterr = ret;
			}
		}
	}
	return lasterr;
}


void usage()
{
	const char *msg = 
		COPYRIGHT
		"\n\n"
		"usage: touch [<option>]... <file>...\n"
		" <option>\n"
		"  -a           archive file (NOT IMPLEMENTED)\n"
		"  -c           no create\n"
		"  -A           change the access time\n"
		"  -C           change the creation time\n"
		"  -M           change the modification time\n"
		"  -R           search subdirectories recursively\n"
		"  -d <string>  parse <string> use it instead of current time\n"
		"  -t <stamp>   use [[CC]YY]MMDDhhmm[.ss[lll]] instead of current time\n"
		"  -D <date>    use [[CC]YY-]MM-DD instead of current date\n"
		"  -T <time>    use hh:mm[:ss[.lll]] instead of current time\n"
		"  -r <rfile>   use <rfile>'s times instead of current time\n"
		"  -F           FAT time stamp\n"
		"  -h,-H,-?     show this message\n"
		"\n"
		;
	
	printf("%s", msg);
	exit(0);
}


/* -r <rfile> */
int parse_reference(LPCTSTR str, SYSTEMTIME *lt)
{
	FILETIME ft;
	BOOL ret;
	TCHAR buf[MAX_PATH];
	
	if (str == NULL) {
		usage();
	}
	
	_tcscpy(buf, str);
	PathRemoveBackslash(buf);
	ret = GetDirectoryTime(buf, NULL, NULL, &ft);
	if (!ret) {
		_ftprintf(stderr, _T("Can't get the file time: %s\n"), buf);
		exit(1);
	}
	ktFileTimeToLocalTime(&ft, lt);
	return 0;
}

/* -d <string> */
int parse_gnu_date(LPCTSTR str, SYSTEMTIME *lt)
{
	TCHAR buf[256];
	LPTSTR p, delim = _T(" ");
	int time = 0;
	
	_tcscpy(buf, str);
	p = _tcstok(buf, delim);
	while (p != NULL) {
		if (_tcschr(p, '-') || _tcschr(p, '/')) {
			parse_date(p, lt);
		} else if (_tcschr(p, ':')) {
			parse_time(p, lt);
			time = 1;
		}
		p = _tcstok(NULL, delim);
	}
	if (!time) {
		lt->wHour = 0;
		lt->wMinute = 0;
		lt->wSecond = 0;
		lt->wMilliseconds = 0;
	}
	return 0;
}

/* -t [[CC]YY]MMDDhhmm[.ss[lll]] */
int parse_gnu_time(LPCTSTR str, SYSTEMTIME *lt)
{
	DWORD n, year, month, day, hour, min, sec = 0, msec = 0;
	TCHAR buf[256] = _T(""), buf2[256];
	LPCTSTR s = buf;
	size_t len;
	int i, ret;
	
	if (str == NULL) {
		usage();
	}
	
	if (_tcschr(str, ':')) {
		return parse_time(str, lt);
	}
	
	ret = _stscanf(str, _T("%[0-9].%5[0-9]"), buf, buf2);
	if (ret == 0) {
		goto error;
	}
	if (ret == 2) {		// .ss[lll]
		len = _tcslen(buf2);
		if (len < 2) {
			goto error;
		}
		n = _ttoi(buf2);
		for (i = 5; i > len; i--) {
			n *= 10;
		}
		sec = n / 1000;
		msec = n % 1000;
	}
	
	len = _tcslen(buf);
	if (len == 10 || len == 12) {		// YYMMDDhhmm or CCYYMMDDhhmm
		_stscanf(buf, _T("%4u"), &year);
		s += len - 8;
		if (len == 10) {				// YYMMDDhhmm
			year /= 100;
			year += (year < 69) ? 2000 : 1900;
		}
		lt->wYear = (WORD) year;
	} else if (len != 8) {
		goto error;
	}
	n = _ttoi(s);
	min = n % 100;
	n /= 100;
	hour = n % 100;
	n /= 100;
	day = n % 100;
	month = n / 100;
	
	lt->wMonth = (WORD) month;
	lt->wDay = (WORD) day;
	lt->wHour = (WORD) hour;
	lt->wMinute = (WORD) min;
	lt->wSecond = (WORD) sec;
	lt->wMilliseconds = (WORD) msec;
	
	return 0;
	
error:
	fprintf(stderr, "Time format error.\n");
	exit(1);
}

/* -D [[CC]YY-]MM-DD, -D [[CC]YY/]MM/DD */
int parse_date(LPCTSTR str, SYSTEMTIME *lt)
{
	DWORD year;
	TCHAR s1[2], s2[2];
	size_t len;
	int ret;
	
	if (str == NULL) {
		usage();
	}
	
	len = _tcslen(str);
	switch (len) {
	case 5:		// MM-DD or MM/DD
		ret = _stscanf(str, _T("%02hu%1[-/]%02hu"), &lt->wMonth, s1, &lt->wDay);
		if (ret < 3) {
			goto error;
		}
		break;
		
	case 8:		// YY-MM-DD or YY/MM/DD
	case 10:	// CCYY-MM-DD or CCYY/MM/DD
		ret = _stscanf(str, _T("%u%1[-/]%02hu%1[-/]%02hu"),
				&year, s1, &lt->wMonth, s2, &lt->wDay);
		if (ret < 5 || (*s1 != *s2)) {
			goto error;
		}
		if (len == 8) {		// YY-MM-DD or YY/MM/DD
			year += (year < 69) ? 2000 : 1900;
		}
		lt->wYear = (WORD) year;
		break;
		
	default:
		goto error;
	}
	return 0;
	
error:
	fprintf(stderr, "Date format error.\n");
	exit(1);
}

/* -T hh:mm[:ss[.lll]] */
int parse_time(LPCTSTR str, SYSTEMTIME *lt)
{
	DWORD sec = 0, msec = 0;
	TCHAR buf[256];
	size_t len;
	int ret, i;
	
	if (str == NULL) {
		usage();
	}
	
	ret = _stscanf(str, _T("%02hu:%02hu:%02u.%3[0-9]"),
			&lt->wHour, &lt->wMinute, &sec, buf);
	if (ret < 2) {
		fprintf(stderr, "Time format error.\n");
		exit(1);
	}
	if (ret == 4) {		// hh:mm:ss.lll
		len = _tcslen(buf);
		msec = _ttoi(buf);
		for (i = 3; i > len; i--) {
			msec *= 10;
		}
	}
	lt->wSecond = (WORD) sec;
	lt->wMilliseconds = (WORD) msec;
	
	return 0;
}


#ifdef DEBUG
void DispFileTime(const FILETIME *ft)
{
	SYSTEMTIME lt;
	
	if (ft == NULL) {
		printf("(null)");
		return;
	}
	
	ktFileTimeToLocalTime(ft, &lt);
	printf("%04d/%02d/%02d %02d:%02d:%02d.%03d",
			lt.wYear, lt.wMonth, lt.wDay, lt.wHour, lt.wMinute, lt.wSecond,
			lt.wMilliseconds);
}
#endif


int touch(LPCTSTR path, const FILETIME *ftCreate,
		const FILETIME *ftAccess, const FILETIME *ftModify,
		int nocreate, int recursive, int fattime)
{
	TCHAR buf[MAX_PATH];
	LPTSTR fname;
	
#ifdef DEBUG
	_tprintf(_T("path: %s\n"), path);
	printf("access: ");
	DispFileTime(ftAccess);
	printf("\n");
	printf("create: ");
	DispFileTime(ftCreate);
	printf("\n");
	printf("modify: ");
	DispFileTime(ftModify);
	printf("\n");
	printf("nocreate: %d\n", nocreate);
	printf("recursive: %d\n", recursive);
#endif
	
	if ((_tcschr(path, '*') == NULL) && (_tcschr(path, '?') == NULL)) {
		/* no wild cards */
		if (!PathFileExists(path)) {
			if (!nocreate) {
				int ret = create_new(path, ftCreate, ftAccess, ftModify);
				if (ret) {
					_ftprintf(stderr, _T("touch error: %s\n"), path);
				}
				return ret;
			}
			return 0;
		} else {
			BOOL ret;
			
#ifdef DEBUG
_tprintf(_T("%s\n"), path);
#endif
			ret = SetFTime(path, ftCreate, ftAccess, ftModify, fattime);
			if (!ret) {
				_ftprintf(stderr, _T("touch error: %s\n"), path);
			}
			if (recursive && PathIsDirectory(path)) {
				return touch_subdir(path, _T("*"), ftCreate, ftAccess, ftModify,
						recursive, fattime);
			}
			return !ret;
		}
	}
	
	fname = PathFindFileName(path);
	_tcsncpy(buf, path, fname - path);
	buf[fname - path] = '\0';
	
	return touch_subdir(buf, fname, ftCreate, ftAccess, ftModify, recursive,
			fattime);
}


int create_new(LPCTSTR path, const FILETIME *ftCreate,
		const FILETIME *ftAccess, const FILETIME *ftModify)
{
	HANDLE hFile;
	BOOL ret;
	
	hFile = CreateFile(path, GENERIC_WRITE, FILE_SHARE_READ, NULL,
			OPEN_ALWAYS /* CREATE_NEW */, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hFile == INVALID_HANDLE_VALUE) {
		return 1;
	}
	ret = SetFileTime(hFile, ftCreate, ftAccess, ftModify);
	CloseHandle(hFile);
	return !ret;
}


int touch_subdir(LPCTSTR path, LPCTSTR file, const FILETIME *ftCreate,
		const FILETIME *ftAccess, const FILETIME *ftModify,
		int recursive, int fattime)
{
	HANDLE hFind;
	BOOL ret = TRUE;
	WIN32_FIND_DATA wfd;
	TCHAR buf[MAX_PATH], buf2[MAX_PATH];
	
	_tcscpy(buf, path);
	ret = PathAppend(buf, _T("*"));
	
	hFind = FindFirstFile(buf, &wfd);
	if (hFind == INVALID_HANDLE_VALUE) {
		return 1;
	}
	do {
		if ((_tcscmp(wfd.cFileName, _T(".")) != 0)
				&& (_tcscmp(wfd.cFileName, _T("..")) != 0)) {
			_tcscpy(buf2, path);
			ret = PathAppend(buf2, wfd.cFileName);
			if (PathMatchSpec(wfd.cFileName, file)) {
				ret = SetFTime(buf2, ftCreate, ftAccess, ftModify, fattime);
				if (!ret) {
					_ftprintf(stderr, _T("touch error: %s\n"), buf2);
				}
#ifdef DEBUG
_tprintf(("%s\n"), buf2);
#endif
			}
			if (recursive
					&& (wfd.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)) {
				ret = !touch_subdir(buf2, file, ftCreate, ftAccess, ftModify,
						recursive, fattime);
			}
		}
	} while (FindNextFile(hFind, &wfd) /*&& ret*/);
	FindClose(hFind);
	
	return !ret;
}


BOOL SetFTime(LPCTSTR path, const FILETIME *pCreationTime,
		const FILETIME *pLastAccessTime, const FILETIME *pLastWriteTime,
		BOOL fattime)
{
	BOOL ret;
	FILETIME ft1, ft2, ft3;
	SYSTEMTIME lt;
	
	
	if (!fattime) {
		return SetDirectoryTime(path, pCreationTime, pLastAccessTime,
				pLastWriteTime);
	}
	
	
	/* NTFS to FAT time stamp */
	
	ret = GetDirectoryTime(path, &ft1, &ft2, &ft3);
	if (!ret) {
		return FALSE;
	}
	
	// creation time precision: 10ms
	((ULARGE_INTEGER *) &ft1)->QuadPart =
			(((ULARGE_INTEGER *) &ft1)->QuadPart + 100000 - 1)
				/ 100000 * 100000;
	
	// access time precision: 1day
	ktFileTimeToLocalTime(&ft2, &lt);
	lt.wHour = 0;
	lt.wMinute = 0;
	lt.wSecond = 0;
	lt.wMilliseconds = 0;
	ktLocalTimeToFileTime(&lt, &ft2);
	
	// modification time precision: 2sec
	((ULARGE_INTEGER *) &ft3)->QuadPart =
			(((ULARGE_INTEGER *) &ft3)->QuadPart + 20000000 - 1)
				/ 20000000 * 20000000;
	
	return SetDirectoryTime(path, &ft1, &ft2, &ft3);
}
