/*
 *	vwinfunc.c - VWin32 functions
 */

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include "vwinfunc.h"


static DWORD dwVWin32Ver = 0;
static DWORD dwVWin32Int21 = VWIN32_Int21Dispatch;
static DWORD dwVWin32Int31 = VWIN32_Int31Dispatch;


/* Get Version and Initialize */
DWORD VWin32GetVersion(void)
{
	if (dwVWin32Ver == 0) {
		DWORD ver = (IsWindows9x()) ? VxDCall0(VWin32_Get_Version) : -1;
		
		if ((0x0104 <= ver) && (ver < 0x040A)) {
			dwVWin32Int21 = VWIN32_Int21Dispatch_;
			dwVWin32Int31 = VWIN32_Int31Dispatch_;
		}
		dwVWin32Ver = ver;
	}
	return dwVWin32Ver;
}


/* Execute Protected Mode Int 21h */
void VWin32PMInt21(DPMIREGS *regs)
{
//	DWORD regEAX, regECX;
//	int rc = 0;
	
	if (dwVWin32Ver == 0) {
		VWin32GetVersion();
	}
	
	__asm {
		pushad
		mov		eax, [regs]
		mov		edi, [eax]		// regs->edi
		mov		esi, [eax+04h]	// regs->esi
	//	mov		ebp, [eax+08h]	// regs->ebp
		mov		ebx, [eax+10h]	// regs->ebx
		mov		edx, [eax+14h]	// regs->edx
		
		// VxDCall2(dwVWin32Int21, regs->eax, regs->ecx);
		mov		ecx, [eax+18h]	// regs->ecx
		push	ecx
		mov		eax, [eax+1Ch]	// regs->eax
		push	eax
		mov		eax, [dwVWin32Int21]
		push	eax
		call	dword ptr VxDCall2
		
		pushfd
		push	eax
		mov		eax, [regs]
		mov		[eax], edi		// regs->edi
		mov		[eax+04h], esi	// regs->esi
	//	mov		[eax+08h], ebp	// regs->ebp
		mov		[eax+10h], ebx	// regs->ebx
		mov		[eax+14h], edx	// regs->edx
		mov		[eax+18h], ecx	// regs->ecx
		pop		ecx
		mov		[eax+1Ch], ecx	// regs->eax
		pop		ecx
		mov		[eax+20h], cx	// regs->flags
		popad
	}
//	return rc;
}


/* Simulate Real Mode Interrupt */
DWORD VWin32RMInt(int intno, DPMIREGS *regs)
{
	DWORD regEAX, regECX;
	int rc = 0;
	
	if (dwVWin32Ver == 0) {
		VWin32GetVersion();
	}
	
	__asm {
		pushad
	
	// DPMI #0300h: Simulate Real Mode Interrupt
	// AX = 0300H
	// BL = interrupt number
	// BH = flags (should be 0)
	// CX = # of words to copy from protected mode to real mode stack
	// ES:(E)DI = selector:offset of real mode register data structure
		mov		ebx, [intno]	// BL=int-no, BH=flags(0)
		mov		edi, [regs]		// register buffer
	}
	regEAX = 0x0300;
	regECX = 0;
	VxDCall2(dwVWin32Int31, regEAX, regECX);
	__asm {
		jnc		Done			// jump when success
		movzx	eax, ax			// AX = DPMI error code (DPMI 1.0+)
		or		eax, 80000000h
		mov		[rc], eax
	Done:
		popad
	}
	return rc;
}


#ifndef NO_ISWIN9X
BOOL IsWindows9x(void)
{
	static DWORD dwVersion = 0;
	static BOOL IsWindows9x = FALSE;
	
	if (dwVersion == 0) {
		DWORD ver = GetVersion();
		IsWindows9x = ((ver & 0x80000000) != 0);
		dwVersion = ver;
	}
	return IsWindows9x;
}
#endif /* NO_ISWIN9X */
